﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Xunit;
using System.IO;
using Genuilder.Tests;

namespace Genuilder.Extensibility.Tests
{
	public class SpyExtension : IExtension
	{
		#region IExtension Members

		public void Execute(ExtensionContext extensionContext)
		{
			if(DeleteFile != null)
				extensionContext.GenItems.First(i => i.Name == DeleteFile).Delete();

			foreach(var item in extensionContext.GenItems)
			{
				foreach(var addChild in _AddChild)
				{
					if(addChild.Item1 == item.Name)
						item.CreateChildItem(addChild.Item2);
				}
			}
			if(CreateFile != null)
				extensionContext.CreateChildItem(CreateFile);
			GenItems = extensionContext.GenItems;
			if(Throw != null)
				throw Throw;
		}

		#endregion

		public GenItemCollection GenItems
		{
			get;
			private set;
		}

		public string CreateFile
		{
			get;
			set;
		}

		public Exception Throw
		{
			get;
			set;
		}

		public string DeleteFile
		{
			get;
			set;
		}

		List<Tuple<string, string>> _AddChild = new List<Tuple<string, string>>();
		public void AddChild(string parent, string child)
		{
			_AddChild.Add(Tuple.Create(parent, child));
		}
	}
	public class TestClass
	{
		[Fact]
		public void CanEnumerateCompileInputFile()
		{
			GenEngine engine = new GenEngine();
			engine.SourceFiles.AddRange(new string[] { "Test.cs", "Test1.cs", "Test2.cs" });
			var spy = engine.Extensions.Add(new SpyExtension());
			engine.Run();
			spy.GenItems.AssertHas("Test.cs", "Test1.cs", "Test2.cs");
		}

		[Fact]
		public void CanSeeModifiedFiles()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test.cs", "Test1.cs", "Test2.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			var result = engine.Run();
			Modify("Test1.cs");

			engine = new GenEngine();
			engine.SourceFiles.AddRange(files);
			engine.StartInfo.RunHistory = result.History;
			spy = engine.Extensions.Add(new SpyExtension());
			engine.Run();
			spy.GenItems.AssertHas(i => i.Modified && i.File.Name == "Test1.cs");
			spy.GenItems.AssertHas(i => !i.Modified && i.File.Name == "Test2.cs");
			spy.GenItems.AssertHas(i => !i.Modified && i.File.Name == "Test.cs");
		}

		[Fact]
		public void CanModifyGenerationDirectory()
		{
			using(var directory = new TestDirectory())
			{
				GenEngine engine = new GenEngine();
				engine.StartInfo.GenerationDirectory = directory.Path;
				var spy = engine.Extensions.Add(new SpyExtension());
				spy.CreateFile = "Hello";
				var spy2 = engine.Extensions.Add(new SpyExtension());
				engine.Run();
				spy2.GenItems.AssertHas(i => i.Name == directory.Path + "\\Hello");
			}
		}

		[Fact]
		public void CanDeleteGenItem()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test.cs", "Test1.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.DeleteFile = "Test1.cs";
			var spy2 = engine.Extensions.Add(new SpyExtension());
			engine.Run().AssertNoError();
			spy2.GenItems.AssertHasNot(i => i.File.Name == "Test1.cs");
			spy2.GenItems.AssertHas(i => i.File.Name == "Test.cs");
		}

		[Fact]
		public void CanAddChildren()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test.cs", "Test1.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.AddChild("Test1.cs", "Test11.cs");
			spy.AddChild("Test1.cs", "Test12.cs");
			var spy2 = engine.Extensions.Add(new SpyExtension());
			engine.Run().AssertNoError();
			spy2.GenItems.AssertHas(i => i.File.Name == "Test.cs");
			spy2.GenItems.AssertHas(i => i.File.Name == "Test11.cs");
			spy2.GenItems.AssertHas(i => i.File.Name == "Test12.cs");
		}


		[Fact]
		public void CanAddRootGenItem()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test.cs", "Test1.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.CreateFile = "Test2.cs";
			var spy2 = engine.Extensions.Add(new SpyExtension());
			engine.Run().AssertNoError();
			spy2.GenItems.AssertHas(i => i.File.Name == "Test.cs");
			spy2.GenItems.AssertHas(i => i.File.Name == "Test1.cs");
			spy2.GenItems.AssertHas(i => i.File.Name == "Test2.cs");
		}

		[Fact]
		public void DeletingParentDeleteChildren()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test1.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.AddChild("Test1.cs", "Test11.cs");
			spy.AddChild("Test1.cs", "Test12.cs");
			var spy2 = engine.Extensions.Add(new SpyExtension());
			spy2.DeleteFile = "Test1.cs";
			var spy3 = engine.Extensions.Add(new SpyExtension());

			engine.Run().AssertNoError();
			Assert.Empty(spy3.GenItems);
		}

		[Fact]
		public void GeneratedSourceAreKeptBetweenTwoRun()
		{
			GenEngine engine = new GenEngine();
			var files = CreateFiles("Test1.cs");
			engine.SourceFiles.AddRange(files);
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.AddChild("Test1.cs", "Test11.cs");
			spy.AddChild("Test1.cs", "Test12.cs");
			var spy2 = engine.Extensions.Add(new SpyExtension());
			var result = engine.Run();

			engine = new GenEngine();
			engine.SourceFiles.AddRange(files);
			engine.StartInfo.RunHistory = result.History;
			spy = engine.Extensions.Add(new SpyExtension());
			spy2 = engine.Extensions.Add(new SpyExtension());
			engine.Run();
			spy.GenItems.AssertHas("Test11.cs", "Test12.cs");
			spy.GenItems.AssertHas(o => o.File.Name == "Test11.cs" && o.Generated);
			spy.GenItems.AssertHas(o => o.File.Name == "Test1.cs" && !o.Generated);
		}

		[Fact]
		public void ExtensionsCanThrow()
		{
			GenEngine engine = new GenEngine();
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.Throw = new Exception();
			var result = engine.Run();
			Assert.NotNull(result.Error);
			Assert.Equal(spy, result.Error.ExtensionContext.Extension);
			Assert.Equal(spy.Throw, result.Error.Exception);
		}

		[Fact]
		public void CanCreateGenItemTwice()
		{
			GenEngine engine = new GenEngine();
			var spy = engine.Extensions.Add(new SpyExtension());
			spy.CreateFile = "File.cs";
			var result = engine.Run();
			engine.StartInfo.RunHistory = result.History;
			engine.Run(true);
		}

		private void Modify(string file)
		{
			File.WriteAllText(file, "hello");
		}

		private string[] CreateFiles(params string[] files)
		{
			foreach(var file in files)
			{
				if(File.Exists(file))
					File.Delete(file);
				File.Create(file).Close();
			}
			return files;
		}
	}
}
