﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Build.Utilities;
using Microsoft.Build.Framework;
using System.IO;
using System.Reflection;

namespace Genuilder.Extensibility.MSBuild
{

	public class ItemProperties
	{
		ExtensibilityTask _Task;
		public ItemProperties(ExtensibilityTask task, SourceType sourceType)
		{
			_Task = task;
			var itemName = sourceType.ToString();
			_SourceType = sourceType;
			_ItemName = itemName;
			_ItemIn = GetProperty(itemName, "In", false);
			_ItemAdded = GetProperty(itemName, "Added", true);
			_ItemRemoved = GetProperty(itemName, "Removed", true);
		}

		private readonly string _ItemName;
		public string ItemName
		{
			get
			{
				return _ItemName;
			}
		}

		private PropertyInfo GetProperty(string itemName, string type, bool isOutput)
		{
			var propertyName = itemName + type;
			var property = typeof(ExtensibilityTask).GetProperty(propertyName);
			if(property == null)
				throw new InvalidOperationException(propertyName + " should exist on ExtensibilityTask but that is not the case");
			if(isOutput)
				if(!property.IsDefined(typeof(OutputAttribute), true))
					throw new InvalidOperationException(propertyName + " should have output attribute");
			return property;
		}

		private readonly SourceType _SourceType;
		public SourceType SourceType
		{
			get
			{
				return _SourceType;
			}
		}

		private readonly PropertyInfo _ItemIn;
		public ITaskItem[] ItemIn
		{
			get
			{
				return GetProperty(_ItemIn);
			}
			set
			{
				SetProperty(_ItemIn, value);
			}
		}

		private ITaskItem[] GetProperty(PropertyInfo property)
		{
			return (ITaskItem[])property.GetValue(_Task, null);
		}

		private void SetProperty(PropertyInfo property, ITaskItem[] value)
		{
			property.SetValue(_Task, value, null);
		}
		private readonly PropertyInfo _ItemRemoved;
		public ITaskItem[] ItemRemoved
		{
			get
			{
				return GetProperty(_ItemRemoved);
			}
			set
			{
				SetProperty(_ItemRemoved, value);
			}
		}
		private readonly PropertyInfo _ItemAdded;
		public ITaskItem[] ItemAdded
		{
			get
			{
				return GetProperty(_ItemAdded);
			}
			set
			{
				SetProperty(_ItemAdded, value);
			}
		}


		List<SourceFile> _SourceFileIn;
		public List<SourceFile> SourceFileIn
		{
			get
			{
				if(_SourceFileIn == null)
				{
					_SourceFileIn = ItemIn.Select(f => new SourceFile(f.ItemSpec, SourceType, GetMetadata(f))).ToList();
				}
				return _SourceFileIn;
			}
		}

		private SourceMetadataCollection GetMetadata(ITaskItem item)
		{
			SourceMetadataCollection collection = new SourceMetadataCollection();
			foreach(var name in item.MetadataNames.OfType<string>())
			{
				collection[name] = item.GetMetadata(name);
			}
			return collection;
		}
	}
	public class ExtensibilityTask : AppDomainIsolatedTask
	{
		public ITaskItem[] CompileIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] CompileRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] CompileAdded
		{
			get;
			set;
		}
		public ITaskItem[] NoneIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] NoneRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] NoneAdded
		{
			get;
			set;
		}

		public ITaskItem[] PageIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] PageRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] PageAdded
		{
			get;
			set;
		}

		public ITaskItem[] ApplicationDefinitionIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] ApplicationDefinitionRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] ApplicationDefinitionAdded
		{
			get;
			set;
		}


		public ITaskItem[] EntityDeployIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] EntityDeployRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] EntityDeployAdded
		{
			get;
			set;
		}

		public ITaskItem[] EmbeddedResourceIn
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] EmbeddedResourceRemoved
		{
			get;
			set;
		}
		[Output]
		public ITaskItem[] EmbeddedResourceAdded
		{
			get;
			set;
		}

		public String CurrentProject
		{
			get;
			set;
		}
		public String OutputPath
		{
			get;
			set;
		}

		public ITaskItem[] Extensions
		{
			get;
			set;
		}


		public string BuildingInsideVisualStudio
		{
			get;
			set;
		}


		public override bool Execute()
		{
			//When a xaml file reference a local type, it compiles a temporary assembly. If a temporary assembly is generated, Genuilder already runned.
			bool isGeneratingTemporaryTargetAssembly = CurrentProject.EndsWith("tmp_proj");
			if(isGeneratingTemporaryTargetAssembly)
				return true;

			if(Extensions == null)
				return true;

			var itemPropertiesSets = GetItemProperties();
			foreach(var itemProperties in itemPropertiesSets)
			{
				if(itemProperties.ItemIn == null)
					itemProperties.ItemIn = new ITaskItem[0];
			}

			var historyPath = Path.Combine(OutputPath, "RunHistory.xml");
			GenEngine engine = new GenEngine();
			engine.Logger = new MSBuildLogger(this.Log);
			engine.BuildingInsideVisualStudio = (BuildingInsideVisualStudio == "true");
			engine.StartInfo.GenerationDirectory = OutputPath;
			engine.StartInfo.Project = CurrentProject;

			engine.SourceFiles.AddRange(itemPropertiesSets.SelectMany(i => i.SourceFileIn));
			try
			{
				engine.Extensions.AddRange(GetExtensions());
			}
			catch(FileNotFoundException ex)
			{
				Log.LogError(ex.Message);
				return false;
			}
			engine.StartInfo.RunHistory = RunHistory.LoadOrNull(historyPath);
			var result = engine.Run(true);
			if(Log.HasLoggedErrors)
				return false;
			result.History.Save(historyPath);

			var sources = result.ExecutionContext.SourceFiles.ToList();
			foreach(var itemProperties in itemPropertiesSets)
			{
				itemProperties.ItemAdded = sources
					.Where(s => s.SourceType == itemProperties.SourceType)
					.Where(s => !itemProperties.SourceFileIn.Any(t => t.Name == s.Name))
					.Select(s => new TaskItem(s.Name))
					.ToArray();

				itemProperties.ItemRemoved = itemProperties.SourceFileIn
					.Where(s => !sources.Where(t => t.SourceType == itemProperties.SourceType).Any(t => t.Name == s.Name))
					.Select(s => new TaskItem(s.Name))
					.ToArray();
			}
			return true;
		}

		private IEnumerable<ItemProperties> GetItemProperties()
		{
			foreach(var value in Enum.GetValues(typeof(SourceType)))
			{
				yield return new ItemProperties(this, (SourceType)value);
			}
		}

		private TaskItem CreateTaskItem(string sourceFile)
		{
			var item = new TaskItem(sourceFile);
			item.SetMetadata("Visible", "false");
			return item;
		}
		private List<IExtension> GetExtensions()
		{
			Dictionary<string, Assembly> loadedAssemblies = new Dictionary<string, Assembly>();
			return Extensions.Select(e => e.ItemSpec.ToExtension(loadedAssemblies)).ToList();
		}
	}
}
